"""
Code Adapted from TabArena: https://github.com/autogluon/tabarena/blob/main/tabarena/tabarena/benchmark/models/ag/tabicl/tabicl_model.py
"""

from __future__ import annotations

import logging

import pandas as pd

from autogluon.common.utils.pandas_utils import get_approximate_df_mem_usage
from autogluon.common.utils.resource_utils import ResourceManager
from autogluon.tabular import __version__
from autogluon.tabular.models.abstract.abstract_torch_model import AbstractTorchModel

logger = logging.getLogger(__name__)


# TODO: Verify if crashes when weights are not yet downloaded and fit in parallel
class TabICLModel(AbstractTorchModel):
    """
    TabICL is a foundation model for tabular data using in-context learning
    that is scalable to larger datasets than TabPFNv2. It is pretrained purely on synthetic data.

    The default TabICL version used is TabICLv2.

    TabICL is one of the top performing methods overall on TabArena-v0.1: https://tabarena.ai
    TabICLv2 significantly improves upon TabICLv1, and achieves very strong performance on TabArena.

    Paper: TabICL: A Tabular Foundation Model for In-Context Learning on Large Data
    Authors: Jingang Qu, David Holzmüller, Gaël Varoquaux, Marine Le Morvan

    Paper: TabICLv2: A better, faster, scalable, and open tabular foundation model
    Authors: Jingang Qu, David Holzmüller, Gaël Varoquaux, Marine Le Morvan

    Codebase: https://github.com/soda-inria/tabicl
    License: BSD-3-Clause

    .. versionadded:: 1.4.0
    """

    ag_key = "TABICL"
    ag_name = "TabICL"

    default_classification_model: str | None = "tabicl-classifier-v2-20260212.ckpt"
    default_regression_model: str | None = "tabicl-regressor-v2-20260212.ckpt"

    ag_priority = 65
    seed_name = "random_state"

    def get_model_cls(self):
        if self.problem_type in ["binary", "multiclass"]:
            from tabicl import TabICLClassifier

            model_cls = TabICLClassifier
        else:
            from tabicl import TabICLRegressor

            model_cls = TabICLRegressor
        return model_cls

    @staticmethod
    def _get_batch_size(n_cells: int):
        if n_cells <= 4_000_000:
            return 8
        elif n_cells <= 6_000_000:
            return 4
        else:
            return 2

    def get_checkpoint_version(self, hyperparameter: dict) -> str:
        clf_checkpoint = self.default_classification_model
        reg_checkpoint = self.default_regression_model

        # Resolve HPO
        if "checkpoint_version" in hyperparameter:
            if isinstance(hyperparameter["checkpoint_version"], str):
                clf_checkpoint = hyperparameter["checkpoint_version"]
                reg_checkpoint = hyperparameter["checkpoint_version"]
            elif isinstance(hyperparameter["checkpoint_version"], tuple):
                clf_checkpoint = hyperparameter["checkpoint_version"][0]
                reg_checkpoint = hyperparameter["checkpoint_version"][1]
            else:
                raise ValueError(
                    "checkpoint_version hyperparameter must be either a string or a tuple of two strings (clf, reg)."
                )

        if self.problem_type in ["binary", "multiclass"]:
            return clf_checkpoint

        return reg_checkpoint

    def _fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        num_cpus: int = 1,
        num_gpus: int = 0,
        **kwargs,
    ):
        try:
            import tabicl
        except ImportError as err:
            logger.log(
                40,
                f"\tFailed to import tabicl! To use the TabICL model, "
                f"do: `pip install autogluon.tabular[tabicl]=={__version__}`.",
            )
            raise err

        from torch.cuda import is_available

        device = "cuda" if num_gpus != 0 else "cpu"
        if (device == "cuda") and (not is_available()):
            # FIXME: warn instead and switch to CPU.
            raise AssertionError(
                "Fit specified to use GPU, but CUDA is not available on this machine. "
                "Please switch to CPU usage instead.",
            )

        model_cls = self.get_model_cls()
        hyp = self._get_model_params()
        hyp["batch_size"] = hyp.get("batch_size", self._get_batch_size(X.shape[0] * X.shape[1]))
        self.model = model_cls(
            **hyp,
            device=device,
            n_jobs=num_cpus,
        )
        X = self.preprocess(X, y=y)
        self.model = self.model.fit(
            X=X,
            y=y,
        )

    def get_device(self) -> str:
        return self.model.device_.type

    # TODO: Better to have an official TabICL method for this
    def _set_device(self, device: str):
        device = self.to_torch_device(device)
        self.model.device_ = device
        self.model.device = self.model.device_.type
        self.model.model_ = self.model.model_.to(self.model.device_)
        self.model.inference_config_.COL_CONFIG.device = self.model.device_
        self.model.inference_config_.ROW_CONFIG.device = self.model.device_
        self.model.inference_config_.ICL_CONFIG.device = self.model.device_

    def _get_default_auxiliary_params(self) -> dict:
        default_auxiliary_params = super()._get_default_auxiliary_params()
        default_auxiliary_params.update(
            {
                # TODO: Instead of caps, should we subsample for large datasets?
                "max_rows": 1000000,  # TODO: What should be the cap? 1 million rows works, but unsure if it is good
                "max_features": 2000,  # TODO: What should be the cap? 10k features works, but unsure if it is good
                "max_batch_size": 1024,  # avoid excessive VRAM usage
            }
        )
        return default_auxiliary_params

    @classmethod
    def supported_problem_types(cls) -> list[str] | None:
        return ["binary", "multiclass", "regression"]

    def _get_default_resources(self) -> tuple[int, int]:
        # Use only physical cores for better performance based on benchmarks
        num_cpus = ResourceManager.get_cpu_count(only_physical_cores=True)

        num_gpus = min(1, ResourceManager.get_gpu_count_torch(cuda_only=True))
        return num_cpus, num_gpus

    def _estimate_memory_usage(self, X: pd.DataFrame, **kwargs) -> int:
        hyperparameters = self._get_model_params()
        return self.estimate_memory_usage_static(
            X=X,
            problem_type=self.problem_type,
            num_classes=self.num_classes,
            hyperparameters=hyperparameters,
            **kwargs,
        )

    @classmethod
    def _estimate_memory_usage_static(
        cls,
        *,
        X: pd.DataFrame,
        hyperparameters: dict = None,
        **kwargs,
    ) -> int:
        """
        Heuristic memory estimate that is very primitive.
        Can be vastly improved.
        """
        if hyperparameters is None:
            hyperparameters = {}

        dataset_size_mem_est = 3 * get_approximate_df_mem_usage(X).sum()  # roughly 3x DataFrame memory size
        baseline_overhead_mem_est = 1e9  # 1 GB generic overhead

        n_rows = X.shape[0]
        n_features = X.shape[1]
        batch_size = hyperparameters.get("batch_size", cls._get_batch_size(X.shape[0] * X.shape[1]))
        embedding_dim = 128
        bytes_per_float = 4
        model_mem_estimate = 2 * batch_size * embedding_dim * bytes_per_float * (4 + n_rows) * n_features

        model_mem_estimate *= 1.3  # add 30% buffer

        # FIXME: Likely this is overly conservative now, figure out more accurate memory estimate for TabICLv2
        #  Early testing shows that cutting this in half is safe.
        # TODO: Observed memory spikes above expected values on large datasets, increasing mem estimate to compensate
        model_mem_estimate *= 2.0  # Note: 1.5 is not large enough, still gets OOM

        mem_estimate = model_mem_estimate + dataset_size_mem_est + baseline_overhead_mem_est

        return mem_estimate

    @classmethod
    def _get_default_ag_args_ensemble(cls, **kwargs) -> dict:
        """
        Set fold_fitting_strategy to sequential_local,
        as parallel folding crashes if model weights aren't pre-downloaded.
        """
        default_ag_args_ensemble = super()._get_default_ag_args_ensemble(**kwargs)
        extra_ag_args_ensemble = {
            # FIXME: If parallel, uses way more memory, seems to behave incorrectly, so we force sequential.
            "fold_fitting_strategy": "sequential_local",
            "refit_folds": True,  # Better to refit the model for faster inference and similar quality as the bag.
        }
        default_ag_args_ensemble.update(extra_ag_args_ensemble)
        return default_ag_args_ensemble

    @classmethod
    def _class_tags(cls) -> dict:
        return {"can_estimate_memory_usage_static": True}

    def _more_tags(self) -> dict:
        return {"can_refit_full": True}



"""Helper functions related to NN architectures"""

import numpy as np

from autogluon.core.constants import REGRESSION


def get_embed_sizes(train_dataset, params, num_categs_per_feature):
    """Returns list of embedding sizes for each categorical variable.
    Selects this adaptively based on training_dataset.
    Note: Assumes there is at least one embed feature.
    """
    max_embedding_dim = params["max_embedding_dim"]
    embed_exponent = params["embed_exponent"]
    size_factor = params["embedding_size_factor"]
    embed_dims = [
        int(size_factor * max(2, min(max_embedding_dim, 1.6 * num_categs_per_feature[i] ** embed_exponent)))
        for i in range(len(num_categs_per_feature))
    ]
    return embed_dims


def infer_y_range(y_vals, y_range_extend):
    """Infers good output range for neural network when used for regression."""
    min_y = float(y_vals.min())
    max_y = float(y_vals.max())
    std_y = np.std(y_vals)
    y_ext = y_range_extend * std_y
    if min_y >= 0:  # infer y must be nonnegative
        min_y = max(0, min_y - y_ext)
    else:
        min_y = min_y - y_ext
    if max_y <= 0:  # infer y must be non-positive
        max_y = min(0, max_y + y_ext)
    else:
        max_y = max_y + y_ext
    return (min_y, max_y)


def get_default_layers(problem_type, num_net_outputs, max_layer_width):
    """Default sizes for NN layers."""
    if problem_type == REGRESSION:
        default_layer_sizes = [
            256,
            128,
        ]  # overall network will have 4 layers. Input layer, 256-unit hidden layer, 128-unit hidden layer, output layer.
    else:
        default_sizes = [256, 128]  # will be scaled adaptively
        # base_size = max(1, min(num_net_outputs, 20)/2.0) # scale layer width based on number of classes
        base_size = max(
            1, min(num_net_outputs, 100) / 50
        )  # TODO: Updated because it improved model quality and made training far faster
        default_layer_sizes = [defaultsize * base_size for defaultsize in default_sizes]
    layer_expansion_factor = 1  # TODO: consider scaling based on num_rows, eg: layer_expansion_factor = 2-np.exp(-max(0,train_dataset.num_examples-10000))
    return [int(min(max_layer_width, layer_expansion_factor * defaultsize)) for defaultsize in default_layer_sizes]


def default_numeric_embed_dim(train_dataset, max_layer_width, first_layer_width):
    """Default embedding dimensionality for numeric features."""
    vector_dim = train_dataset.dataset._data[train_dataset.vectordata_index].shape[
        1
    ]  # total dimensionality of vector features
    prop_vector_features = train_dataset.num_vector_features() / float(
        train_dataset.num_features
    )  # Fraction of features that are numeric
    min_numeric_embed_dim = 32
    max_numeric_embed_dim = max_layer_width
    return int(
        min(
            max_numeric_embed_dim,
            max(min_numeric_embed_dim, first_layer_width * prop_vector_features * np.log10(vector_dim + 10)),
        )
    )



# optimized for <=10000 samples and <=500 features, with a GPU present
hyperparameter_portfolio_zeroshot_2025_small = {
    "REALTABPFN-V2": [
        {
            "ag_args": {"priority": -1},
        },
    ],
    "GBM": [
        {
            "ag_args": {"name_suffix": "_r33", "priority": -2},
            "bagging_fraction": 0.9625293420216,
            "bagging_freq": 1,
            "cat_l2": 0.1236875455555,
            "cat_smooth": 68.8584757332856,
            "extra_trees": False,
            "feature_fraction": 0.6189215809382,
            "lambda_l1": 0.1641757352921,
            "lambda_l2": 0.6937755557881,
            "learning_rate": 0.0154031028561,
            "max_cat_to_onehot": 17,
            "min_data_in_leaf": 1,
            "min_data_per_group": 30,
            "num_leaves": 68,
        },
        {
            "ag_args": {"name_suffix": "_r21", "priority": -16},
            "bagging_fraction": 0.7218730663234,
            "bagging_freq": 1,
            "cat_l2": 0.0296205152578,
            "cat_smooth": 0.0010255271303,
            "extra_trees": False,
            "feature_fraction": 0.4557131604374,
            "lambda_l1": 0.5219704038237,
            "lambda_l2": 0.1070959487853,
            "learning_rate": 0.0055891584996,
            "max_cat_to_onehot": 71,
            "min_data_in_leaf": 50,
            "min_data_per_group": 10,
            "num_leaves": 30,
        },
        {
            "ag_args": {"name_suffix": "_r11", "priority": -19},
            "bagging_fraction": 0.775784726514,
            "bagging_freq": 1,
            "cat_l2": 0.3888471449178,
            "cat_smooth": 0.0057144748021,
            "extra_trees": True,
            "feature_fraction": 0.7732354787904,
            "lambda_l1": 0.2211002452568,
            "lambda_l2": 1.1318405980187,
            "learning_rate": 0.0090151778542,
            "max_cat_to_onehot": 15,
            "min_data_in_leaf": 4,
            "min_data_per_group": 15,
            "num_leaves": 2,
        },
    ],
    "CAT": [
        {
            "ag_args": {"priority": -5},
        },
        {
            "ag_args": {"name_suffix": "_r51", "priority": -10},
            "boosting_type": "Plain",
            "bootstrap_type": "Bernoulli",
            "colsample_bylevel": 0.8771035272558,
            "depth": 7,
            "grow_policy": "SymmetricTree",
            "l2_leaf_reg": 2.0107286863021,
            "leaf_estimation_iterations": 2,
            "learning_rate": 0.0058424016622,
            "max_bin": 254,
            "max_ctr_complexity": 4,
            "model_size_reg": 0.1307400355809,
            "one_hot_max_size": 23,
            "subsample": 0.809527841437,
        },
        {
            "ag_args": {"name_suffix": "_r10", "priority": -12},
            "boosting_type": "Plain",
            "bootstrap_type": "Bernoulli",
            "colsample_bylevel": 0.8994502668431,
            "depth": 6,
            "grow_policy": "Depthwise",
            "l2_leaf_reg": 1.8187025215896,
            "leaf_estimation_iterations": 7,
            "learning_rate": 0.005177304142,
            "max_bin": 254,
            "max_ctr_complexity": 4,
            "model_size_reg": 0.5247386875068,
            "one_hot_max_size": 53,
            "subsample": 0.8705228845742,
        },
        {
            "ag_args": {"name_suffix": "_r24", "priority": -15},
            "boosting_type": "Plain",
            "bootstrap_type": "Bernoulli",
            "colsample_bylevel": 0.8597809376276,
            "depth": 8,
            "grow_policy": "Depthwise",
            "l2_leaf_reg": 0.3628261923976,
            "leaf_estimation_iterations": 5,
            "learning_rate": 0.016851077771,
            "max_bin": 254,
            "max_ctr_complexity": 4,
            "model_size_reg": 0.1253820547902,
            "one_hot_max_size": 20,
            "subsample": 0.8120271122061,
        },
        {
            "ag_args": {"name_suffix": "_r91", "priority": -17},
            "boosting_type": "Plain",
            "bootstrap_type": "Bernoulli",
            "colsample_bylevel": 0.8959275863514,
            "depth": 4,
            "grow_policy": "SymmetricTree",
            "l2_leaf_reg": 0.0026915894253,
            "leaf_estimation_iterations": 12,
            "learning_rate": 0.0475233791203,
            "max_bin": 254,
            "max_ctr_complexity": 5,
            "model_size_reg": 0.1633175256924,
            "one_hot_max_size": 11,
            "subsample": 0.798554178926,
        },
    ],
    "TABM": [
        {
            "ag_args": {"name_suffix": "_r184", "priority": -6},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 864,
            "d_embedding": 24,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0019256819924656217,
            "n_blocks": 3,
            "num_emb_n_bins": 3,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {"name_suffix": "_r69", "priority": -7},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 848,
            "d_embedding": 28,
            "dropout": 0.40215621636031007,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0010413640454559532,
            "n_blocks": 3,
            "num_emb_n_bins": 18,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {"name_suffix": "_r52", "priority": -11},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 1024,
            "d_embedding": 32,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0006297851297842611,
            "n_blocks": 4,
            "num_emb_n_bins": 22,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.06900108498839816,
        },
        {
            "ag_args": {"priority": -13},
        },
        {
            "ag_args": {"name_suffix": "_r191", "priority": -14},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 864,
            "d_embedding": 8,
            "dropout": 0.45321529282058803,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0003781238075322413,
            "n_blocks": 4,
            "num_emb_n_bins": 27,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.01766851962579851,
        },
        {
            "ag_args": {"name_suffix": "_r49", "priority": -20},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 640,
            "d_embedding": 28,
            "dropout": 0.15296207419190627,
            "gradient_clipping_norm": 1.0,
            "lr": 0.002277678490593717,
            "n_blocks": 3,
            "num_emb_n_bins": 48,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0578159148243893,
        },
    ],
    "TABICL": [
        {
            "ag_args": {"priority": -8},
        },
    ],
    "XGB": [
        {
            "ag_args": {"name_suffix": "_r171", "priority": -9},
            "colsample_bylevel": 0.9213705632288,
            "colsample_bynode": 0.6443385965381,
            "enable_categorical": True,
            "grow_policy": "lossguide",
            "learning_rate": 0.0068171645251,
            "max_cat_to_onehot": 8,
            "max_depth": 6,
            "max_leaves": 10,
            "min_child_weight": 0.0507304250576,
            "reg_alpha": 4.2446346389037,
            "reg_lambda": 1.4800570021253,
            "subsample": 0.9656290596647,
        },
        {
            "ag_args": {"name_suffix": "_r40", "priority": -18},
            "colsample_bylevel": 0.6377491713202,
            "colsample_bynode": 0.9237625621103,
            "enable_categorical": True,
            "grow_policy": "lossguide",
            "learning_rate": 0.0112462621131,
            "max_cat_to_onehot": 33,
            "max_depth": 10,
            "max_leaves": 35,
            "min_child_weight": 0.1403464856034,
            "reg_alpha": 3.4960653958503,
            "reg_lambda": 1.3062320805235,
            "subsample": 0.6948898835178,
        },
    ],
    "MITRA": [
        {
            "n_estimators": 1,
            "fine_tune": True,
            "fine_tune_steps": 50,
            "ag.num_gpus": 1,
            "ag_args": {"priority": -21},
        },
    ],
}



from __future__ import annotations

import logging
from typing import Type

import numpy as np
import pandas as pd

from autogluon.features import BulkFeatureGenerator
from autogluon.features.generators.abstract import AbstractFeatureGenerator
from autogluon.features.registry._ag_feature_generator_registry import ag_feature_generator_registry

logger = logging.getLogger(__name__)


def _recursive_expand_prep_param(prep_param: tuple | list[list | tuple]) -> list[tuple]:
    if isinstance(prep_param, list):
        if len(prep_param) == 0:
            param_type = "list"
        elif len(prep_param) == 2:
            if isinstance(prep_param[0], (str, AbstractFeatureGenerator)):
                param_type = "generator"
            else:
                param_type = "list"
        else:
            param_type = "list"
    elif isinstance(prep_param, tuple):
        param_type = "generator"
    else:
        raise ValueError(f"Invalid value for prep_param: {prep_param}")
    if param_type == "list":
        out = []
        for p in prep_param:
            out += _recursive_expand_prep_param(p)
        return out
    elif param_type == "generator":
        return [prep_param]
    else:
        raise ValueError(f"Invalid value for prep_param: {prep_param}")


# FIXME: Why is preprocessing twice as slow per fold when bagging LightGBM??? Need to investigate. Try sequential fold fit
# TODO: Why is `prep_params` a dict instead of a list?
class ModelAgnosticPrepMixin:
    def _estimate_dtypes_after_preprocessing(self, X: pd.DataFrame, **kwargs) -> int:
        prep_params = self._get_ag_params().get("prep_params", None)
        if prep_params is None:
            prep_params = []

        # FIXME: Temporarily simplify for memory calculation
        prep_params = _recursive_expand_prep_param(prep_params)

        X_nunique = X.nunique().values
        n_categorical = X.select_dtypes(exclude=[np.number]).shape[1]
        n_numeric = X.loc[:, X_nunique > 2].select_dtypes(include=[np.number]).shape[1]
        n_binary = (
            X.loc[:, X_nunique <= 2].select_dtypes(include=[np.number]).shape[1]
        )  # NOTE: It can happen that features have less than two unique values if cleaning is applied before the bagging, i.e. Bioresponse

        assert n_numeric + n_categorical + n_binary == X.shape[1]  # NOTE: FOr debugging, to be removed later
        for preprocessor_cls_name, init_params in prep_params:
            if isinstance(preprocessor_cls_name, str):
                prep_cls = ag_feature_generator_registry.key_to_cls(key=preprocessor_cls_name)
            else:
                prep_cls = preprocessor_cls_name
            feature_generator = prep_cls(target_type=self.problem_type, **init_params)

            if hasattr(feature_generator, "estimate_new_dtypes"):
                n_numeric, n_categorical, n_binary = feature_generator.estimate_new_dtypes(
                    n_numeric, n_categorical, n_binary, num_classes=self.num_classes
                )
            else:
                pass  # FIXME: Need to implement

        return n_numeric, n_categorical, n_binary

    def _estimate_memory_usage(self, X: pd.DataFrame, **kwargs) -> int:
        hyperparameters = self._get_model_params()
        n_numeric, n_categorical, n_binary = self._estimate_dtypes_after_preprocessing(X=X, **kwargs)

        if hasattr(self, "_estimate_memory_usage_static_lite"):
            return self._estimate_memory_usage_static_lite(
                num_samples=X.shape[0],
                num_features=n_numeric + n_categorical + n_binary,
                num_bytes_per_cell=4,
                hyperparameters=hyperparameters,
                problem_type=self.problem_type,
                num_classes=self.num_classes,
                **kwargs,
            )

        # TODO: Replace with memory estimation logic based on no. of features instead of dataframe generation
        shape = X.shape[0]
        df_lst = []
        if n_numeric > 0:
            X_estimate = np.random.random(size=[shape, n_numeric]).astype(np.float32)
            X_estimate_numeric = pd.DataFrame(X_estimate)
            df_lst.append(X_estimate_numeric)
        if n_categorical > 0:
            cardinality = int(X.select_dtypes(exclude=[np.number]).nunique().mean())
            X_estimate = np.random.randint(0, cardinality, [shape, n_categorical]).astype("str")
            X_estimate_cat = pd.DataFrame(X_estimate)
            df_lst.append(X_estimate_cat)
        if n_binary > 0:
            X_estimate = np.random.randint(0, 2, [shape, n_binary]).astype(np.int8)
            X_estimate_binary = pd.DataFrame(X_estimate)
            df_lst.append(X_estimate_binary)
        X = pd.concat(df_lst, ignore_index=True, axis=1)

        return self.estimate_memory_usage_static(
            X=X,
            problem_type=self.problem_type,
            num_classes=self.num_classes,
            hyperparameters=hyperparameters,
            **kwargs,
        )

    def _init_preprocessor(
        self,
        preprocessor_cls: Type[AbstractFeatureGenerator] | str,
        init_params: dict | None,
    ) -> AbstractFeatureGenerator:
        if isinstance(preprocessor_cls, str):
            preprocessor_cls = ag_feature_generator_registry.key_to_cls(preprocessor_cls)
        if init_params is None:
            init_params = {}
        _init_params = dict(
            verbosity=0,
            random_state=self.random_seed,  # FIXME: Not a generic param
            target_type=self.problem_type,  # FIXME: Not a generic param
        )
        _init_params.update(**init_params)
        return preprocessor_cls(
            **_init_params,
        )

    def _recursive_init_preprocessors(self, prep_param: tuple | list[list | tuple]):
        if isinstance(prep_param, list):
            if len(prep_param) == 0:
                param_type = "list"
            elif len(prep_param) == 2:
                if isinstance(prep_param[0], (str, AbstractFeatureGenerator)):
                    param_type = "generator"
                else:
                    param_type = "list"
            else:
                param_type = "list"
        elif isinstance(prep_param, tuple):
            param_type = "generator"
        else:
            raise ValueError(f"Invalid value for prep_param: {prep_param}")

        if param_type == "list":
            out = []
            for i, p in enumerate(prep_param):
                out.append(self._recursive_init_preprocessors(p))
            return out
        elif param_type == "generator":
            assert len(prep_param) == 2
            preprocessor_cls = prep_param[0]
            init_params = prep_param[1]
            return self._init_preprocessor(preprocessor_cls=preprocessor_cls, init_params=init_params)
        else:
            raise ValueError(f"Invalid value for prep_param: {prep_param}")

    def get_preprocessor(self) -> AbstractFeatureGenerator | None:
        ag_params = self._get_ag_params()
        prep_params = ag_params.get("prep_params", None)
        passthrough_types = ag_params.get("prep_params.passthrough_types", None)
        if prep_params is None:
            return None
        if not prep_params:
            return None

        preprocessors = self._recursive_init_preprocessors(prep_param=prep_params)
        if len(preprocessors) == 0:
            return None
        if len(preprocessors) == 1 and isinstance(preprocessors[0], AbstractFeatureGenerator):
            return preprocessors[0]
        else:
            preprocessor = BulkFeatureGenerator(
                generators=preprocessors,
                # TODO: "false_recursive" technically can slow down inference, but need to optimize `True` first
                #  Refer to `Bioresponse` dataset where setting to `True` -> 200s fit time vs `false_recursive` -> 1s fit time
                remove_unused_features="false_recursive",
                post_drop_duplicates=True,
                passthrough=True,
                passthrough_types=passthrough_types,
                verbosity=0,
            )
            return preprocessor

    def _preprocess(self, X: pd.DataFrame, y=None, is_train: bool = False, **kwargs):
        if is_train:
            self.preprocessor = self.get_preprocessor()
            if self.preprocessor is not None:
                assert y is not None, (
                    f"y must be specified to fit preprocessors... Likely the inheriting class isn't passing `y` in its `preprocess` call."
                )
                # FIXME: add `post_drop_useless`, example: anneal has many useless features
                feature_metadata_in = self._feature_metadata
                X = self.preprocessor.fit_transform(X, y, feature_metadata_in=feature_metadata_in)
                # FIXME: Nick: This is incorrect because it strips away special dtypes. Need to do this properly by fixing in the preprocessors
                feature_metadata_in = self.preprocessor.feature_metadata
                self._feature_metadata = feature_metadata_in
                self._features_internal = self._feature_metadata.get_features()
        else:
            if self.preprocessor is not None:
                X = self.preprocessor.transform(X)

        return super()._preprocess(X, y=y, is_train=is_train, **kwargs)



from __future__ import annotations

from ..lgb.lgb_model import LGBModel
from .prep_mixin import ModelAgnosticPrepMixin


class PrepLGBModel(ModelAgnosticPrepMixin, LGBModel):
    ag_key = "GBM_PREP"
    ag_name = "LightGBMPrep"

    @classmethod
    def _estimate_memory_usage_static(cls, **kwargs) -> int:
        memory_usage = super()._estimate_memory_usage_static(**kwargs)
        # FIXME: 1.5 runs OOM on kddcup09_appetency fold 2 repeat 0 prep_LightGBM_r49_BAG_L1
        return memory_usage * 2.0  # FIXME: For some reason this underestimates mem usage without this

    @classmethod
    def _estimate_memory_usage_static_lite(cls, **kwargs) -> int:
        memory_usage = super()._estimate_memory_usage_static_lite(**kwargs)
        # FIXME: 1.5 runs OOM on kddcup09_appetency fold 2 repeat 0 prep_LightGBM_r49_BAG_L1
        return memory_usage * 2.0  # FIXME: For some reason this underestimates mem usage without this
